"""
TODO: add comments
"""
from command import Command
from utils import init_tow
from tow import templates
import subprocess
from utils import TOW_VOLUME
import os


class BuildCommand(Command):

    def add_parser(self, subparsers):
        super(BuildCommand, self).add_parser(subparsers)
        parser = subparsers.add_parser("build",
                                       help="process attributes and tamplates path Dockerfile according mapping and run docker build with DOCKER-OPTIONS")
        parser.add_argument("--tow-run", action="store_true",
                            help="patch docker file in order to use configuration in run phase")
        parser.add_argument("--tow-attrs", type=str, default="default",
                            help="specify name of attribute file in attributes folder. Default value: default. Example --tow-attrs prod")
        parser.add_argument("--tow-mapping", type=str, default="mapping",
                            help="specify name of mapping variable in mapping.py. Default value:mapping. Example --tow-mapping prod")

    def command(self, namespace, args):
        """
            This is build command. Prepare workingdir(.tow) and run docker build
        """
        (file_mapping, dockerfile, envs, attrs, workingdir) = init_tow(attributes_name=namespace.tow_attrs,
                                                                       mapping_name=namespace.tow_mapping)
        #  Check if you would like to patch Dockerfile in order to use
        #  reconfiguration on run phase
        if namespace.tow_run:
            (entrypoint, cmd) = dockerfile.find_entrypoint_or_cmd()
            templates.process_template("tow.sh.tmpl",
                                       os.path.join(workingdir, "tow.sh"),
                                       {"entrypoint": entrypoint,
                                        "cmd": cmd, "mapping": file_mapping,
                                        "volume_name": TOW_VOLUME})
            file_mapping.append(("tow.sh", "/tow.sh"))
            dockerfile.replace_entrypoint_or_cmd_by_tow_cmd("sh /tow.sh")
        dockerfile.add_copy(file_mapping)
        dockerfile.save(os.path.join(workingdir, "Dockerfile"))

        try:
            subprocess.call(["docker", "build"] + args + [workingdir])
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                print "ERORR: Please install docker and run tow again"



"""
TODO: add comments
"""
import os
import shutil
from tow.modules import load_module
from tow.dockerfile import Dockerfile
from tow.attrs import process_attrs
from tow import templates
import subprocess
import sys
import json
import collections


TOW_VOLUME = "/tow"


def project_paths():
    current_dir = os.getcwd()
    return (current_dir,
            os.path.join(current_dir, "Dockerfile"),
            os.path.join(current_dir, "mapping.py"),
            os.path.join(current_dir, "templates"),
            os.path.join(current_dir, "files"),
            os.path.join(current_dir, "attributes"))


def prepare_workingdir(workingdir):
    if os.path.exists(workingdir):
        shutil.rmtree(workingdir, ignore_errors=True)
    os.mkdir(workingdir)


def dequote(s):
    """
    If a string has single or double quotes around it, remove them.
    Make sure the pair of quotes match.
    If a matching pair of quotes is not found, return the string unchanged.
    """
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        return s[1:-1]
    return s


def parse_env_arg(env_arg):
    if env_arg:
        env_pair = env_arg.split("=")
        if len(env_pair) < 2:
            return (env_pair[0], os.getenv(env_pair[0], ""))
        else:
            return (env_pair[0], dequote("".join(env_pair[1:])))


def parse_envfile(env_file_name):
    envs = []
    with open(env_file_name, "r") as envfile:
        for envfile_line in envfile.readlines():
            if not envfile_line.strip().startswith("#"):
                envs.append(parse_env_arg(envfile_line.strip()))
    return envs


def get_env_args(args):
    envs = {}

    i = 0
    while i < len(args):
        arg = args[i]
        if arg == "-e":
            i = i + 1
            (env_name, env_var) = parse_env_arg(args[i].strip())
            envs[env_name] = env_var
            i = i + 1
        elif arg == "--env":
            i = i + 1
            while i < len(args) and not args[i].startswith("-"):
                (env_name, env_var) = parse_env_arg(args[i].strip())
                envs[env_name] = env_var
                i = i + 1
        elif arg == "--env-file":
            i = i + 1
            env_file_name = args[i].strip()
            env_vars = parse_envfile(env_file_name)
            envs.update({env_var: env_name for (env_name, env_var) in env_vars})
            i = i + 1
        else:
            i = i + 1
    return envs


def copy_files(workingdir, files_path, mapping):
    for fm in mapping.get("files", []):
        src = fm[0]
        src_file_path = os.path.join(files_path, src)
        if os.path.exists(src_file_path):
            dst_file_path = os.path.join(workingdir, src)
            file_path_dir = os.path.dirname(dst_file_path)
            if not os.path.exists(file_path_dir):
                os.makedirs(file_path_dir)
            shutil.copy2(src_file_path, dst_file_path)
        else:
            print "file %s doesn't exists" % src_file_path


def init_tow(env_args={}, attributes_name="default", mapping_name="mapping"):
    (current_dir, dockerfile_path,
     mappingfile_path, templates_path,
     files_path, attributes_path) = project_paths()

    file_mapping = load_module({}, current_dir, "mapping")
    mapping = getattr(file_mapping, mapping_name, {})

    workingdir = os.path.join(current_dir, ".tow")

    prepare_workingdir(workingdir)

    dockerfile = Dockerfile(dockerfile_path)

    envs = dockerfile.envs()
    # envs passed as params has more priority then Dockerfile envs
    envs.update(env_args)
    attrs = process_attrs(envs, attributes_path, attributes_name)

    # process templates
    for fm in mapping.get("templates", []):
        src = fm[0]
        src_template_path = os.path.join(templates_path, src)
        if os.path.exists(src_template_path):
            processed_template_path = os.path.join(workingdir, src)
            template_path_dir = os.path.dirname(processed_template_path)
            if not os.path.exists(template_path_dir):
                os.makedirs(template_path_dir)

            templates.process(os.path.dirname(src_template_path),
                              os.path.basename(src_template_path),
                              processed_template_path, attrs)
        else:
            print "WARN: template file %s doesn't exists" % src_template_path

    copy_files(workingdir, files_path, mapping)

    # Transform dict with mapping to list of file tuples that exists in .tow dir
    handled_file_mapping = [fm for fm_list in mapping.values()
                            for fm in fm_list
                            if os.path.exists(os.path.join(workingdir, fm[0]))]

    # Init mapping file
    templates.process_template("mapping.sh.tmpl",
                               os.path.join(workingdir, "mapping.sh"),
                               {"mapping": handled_file_mapping,
                                "volume_name": TOW_VOLUME})

    return (handled_file_mapping, dockerfile, envs, attrs, workingdir)


def get_linked_container_variables(args):
    envs = {}
    if "--link" in args:
        link_idx = args.index("--link")
        link_idx = link_idx + 1
        link_info = args[link_idx]
        (name, alias) = link_info.split(":")
        current_name_idx = args.index("--name")
        current_name_idx = current_name_idx + 1
        current_name = args[current_name_idx]

        p = subprocess.Popen(["docker", "inspect", name],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if err:
            print "ERROR: Problem to make docker inspect for %s, %s" % (name, err)
            sys.exit(1)
        linked_info = json.loads(out.strip(), object_pairs_hook=collections.OrderedDict)
        exposed_ports = [port for port in linked_info[0]["Config"]["ExposedPorts"].keys()]
        linked_envs = [env for env in linked_info[0]["Config"]["Env"]
                       if not (env.startswith("HOME") or env.startswith("PATH"))]
        ip_address = linked_info[0]["NetworkSettings"]["IPAddress"]
        envs["%s_NAME" % alias.upper()] = "/%s/%s" % (current_name, alias)
        if exposed_ports:
            firts_port, first_proto = exposed_ports[0].split("/")
            envs["%s_PORT" % alias.upper()] = "%s://%s:%s" % (first_proto,
                                                              ip_address,
                                                              firts_port)
            for exposed_port in exposed_ports:
                port, proto = exposed_port.split("/")
                envs["%s_PORT_%s_%s" % (alias.upper(), port, proto.upper())] = "%s://%s:%s" % (proto,
                                                                                               ip_address,
                                                                                               port)
                envs["%s_PORT_%s_%s_PROTO" % (alias.upper(), port, proto.upper())] = proto
                envs["%s_PORT_%s_%s_PORT" % (alias.upper(), port, proto.upper())] = port
                envs["%s_PORT_%s_%s_ADDR" % (alias.upper(), port, proto.upper())] = ip_address

        if linked_envs:
            for linked_env in linked_envs:
                linked_env_name, linked_env_value = linked_env.split("=")
                envs["%s_ENV_%s" % (alias.upper(), linked_env_name)] = linked_env_value
    return envs



"""
TODO: add comments
"""
from command import Command
from utils import init_tow
from utils import get_env_args
from utils import get_linked_container_variables
import subprocess
import os
from utils import TOW_VOLUME


class RunCommand(Command):

    def add_parser(self, subparsers):
        super(RunCommand, self).add_parser(subparsers)
        parser = subparsers.add_parser("run",
                                       help="run docker container if tow build was without --tow-run option than call docker run else process attributes and tempaltes mount /tow volume and run docker run with DOCKER-OPTIONS")
        parser.add_argument("--tow-attrs", type=str, default="default",
                            help="specify name of attribute file in attributes folder. Default value: default. Example --tow-attrs prod")
        parser.add_argument("--tow-mapping", type=str, default="mapping",
                            help="specify name of mapping variable in mapping.py. Default value:mapping. Example --tow-mapping prod")

    def command(self, namespace, args):
        linked_envs = get_linked_container_variables(args)
        env_args = get_env_args(args)
        env_args.update(linked_envs)
        (file_mapping, dockerfile, envs, attrs, workingdir) = init_tow(env_args=env_args,
                                                                       attributes_name=namespace.tow_attrs,
                                                                       mapping_name=namespace.tow_mapping)

        try:
            subprocess.call(["docker", "run", "-v", "%s:%s" % (workingdir, TOW_VOLUME)] + args)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                print "ERORR: Please install docker and run tow again"



"""
This module contains tow `main` method.

`main` is executed as the command line ``tow`` program and takes care of
parsing options and commands, loading the user settings file
 and executing the commands given.

The other callables defined in this module are internal only.
"""

import sys
from version import version
from argparse import ArgumentParser
from commands.create import CreateCommand
from commands.build import BuildCommand
from commands.run import RunCommand

# don't create pyc files
sys.dont_write_bytecode = True


def main():
    """
    Main command-line execution loop
    """
    tow_parser = ArgumentParser(description="tow is configuration managment tool for docker containers", version=version)
    tow_subparsers = tow_parser.add_subparsers(help="tow commands", dest="command")

    commands = {}

    create_command = CreateCommand()
    create_command.add_parser(tow_subparsers)
    commands["create"] = create_command

    build_command = BuildCommand()
    build_command.add_parser(tow_subparsers)
    commands["build"] = build_command

    run_command = RunCommand()
    run_command.add_parser(tow_subparsers)
    commands["run"] = run_command

    if len(sys.argv) > 1:
        (namespace, args) = tow_parser.parse_known_args()

        if namespace.command not in commands:
                tow_parser.print_help()
        else:
            commands[namespace.command].command(namespace, args)
    else:
        tow_parser.print_help()
    sys.exit(0)



"""
TODO: add comments
"""
from command import Command
from tow import templates
import os
from datetime import date


class CreateCommand(Command):

    def add_parser(self, subparsers):
        super(CreateCommand, self).add_parser(subparsers)
        parser = subparsers.add_parser("create",
                                       help="Create tow project in current directory")
        parser.add_argument("project_name", type=str,
                            help="name of tow project")

    def command(self, namespace, args):
        project_name = namespace.project_name
        for dir_name in ["attributes", "files", "templates"]:
            os.makedirs(os.path.join(project_name, dir_name))
        for file_name in ["Dockerfile", "mapping.py", "attributes/default.py"]:
            templates.process_template("%s.tmpl" % os.path.basename(file_name),
                                    os.path.join(project_name, file_name),
                                    {"current_year": date.today().year,
                                        "project_name": project_name})
